#include "StdAfx.h"
#include "MatrixAlgebra.h"
#include "Quaternion.h"
#include "mm.h"

static Quaternion x;	// EKF  , ʹϾ
static dMatrix P(4,4);	//   л 

double average_gyro[3] = {0, 0, 0, };	// ӵ   Ѵ. ȭ鿡 ǥϱ   .
double adjust_rpy[3] = {0, 0, 0, };		// ӵ  ߷ κ  roll, pitch, yaw . ⼭  .

void InitARS()
{
	// EKF  Ϳ л  ʱȭ Ѵ.
	x.set_s (1);
	x.set_v (dVector3(0, 0, 0));

	P.unit();		// л  ķ ʱȭ
}

void GyroAverage (double average_gyro[3], double gyro[3], const int max_count, const double tolerance)
{
	//   ִ ӵ(gyroscope)   0   ʴ´.
	// κ Ư  ̾ Ǿִµ,   0 ϱ Ͽ
	//     ӵ   Ѵ.

	static int count = 0;

	double d0 = gyro[0] - average_gyro[0];
	double d1 = gyro[1] - average_gyro[1];
	double d2 = gyro[2] - average_gyro[2];

	if (!count || ( 
		-tolerance < d0 && d0 < tolerance && 
		-tolerance < d1 && d1 < tolerance && 
		-tolerance < d2 && d2 < tolerance )) {

		average_gyro[0]  = (average_gyro[0]*count + gyro[0])/(count+1);
		average_gyro[1]  = (average_gyro[1]*count + gyro[1])/(count+1);
		average_gyro[2]  = (average_gyro[2]*count + gyro[2])/(count+1);
	}
	if (count < max_count) ++count;
}

dMatrix GetStateTransitionMatrix(double wx, double wy, double wz, double dt)
{
	dMatrix A(4,4);

	wx *= 0.5*dt;
	wy *= 0.5*dt;
	wz *= 0.5*dt;

	A(0,0) = 1;		A(0,1) = wz;	A(0,2) = -wy;	A(0,3) = wx;
	A(1,0) = -wz;	A(1,1) = 1;		A(1,2) = wx;	A(1,3) = wy;
	A(2,0) = wy;	A(2,1) = -wx;	A(2,2) = 1;		A(2,3) = wz;
	A(3,0) = -wx;	A(3,1) = -wy;	A(3,2) = -wz;	A(3,3) = 1;

	return A;
}


dMatrix GetMeasurementMatrix(Quaternion &q)
{
#define eps		1e-10

	double e1 = q.v_[0];
	double e2 = q.v_[1];
	double e3 = q.v_[2];
	double a = e3*(e1*e1 + e2*e2 + e3*e3);
	double b = sqrt(e1*e1 + e2*e2);

	if (-eps < a && a < eps) a = eps;
	if (b < eps) b = eps;

	const double _h[1][4] = { 
		{ (e1*e3*e3 + eps)/(a*b + eps), (e2*e3*e3 + eps)/(a*b + eps), (e3*b + eps)/(a + eps), 0	},
	};

	// observation matrix
	dMatrix H(1, 4, &_h[0][0]);

	return H;
}

dMatrix ARS (double gyro[3], double accl[3], double magn[3], double dt)
{
	GyroAverage (average_gyro, gyro, 10000, 10.);

	/////////////////////////////////////////////////////////////////////////
	// Predict 

	// ̷  ̾  ϱ  ӵ  ӵ հ .
	dVector w = dVector3 (
		_DEG2RAD*(gyro[0] - average_gyro[0]), 
		_DEG2RAD*(gyro[1] - average_gyro[1]), 
		_DEG2RAD*(gyro[2] - average_gyro[2]));

	// state transition matrix
	dMatrix A = GetStateTransitionMatrix (w[0], w[1], w[2], dt);
	
	double _b[4][3] = {
		{ x.s_,    -x.v_[2],  x.v_[1] },
		{ x.v_[2],  x.s_,    -x.v_[0] },
		{-x.v_[1],  x.v_[0],  x.s_    },
		{-x.v_[0], -x.v_[1], -x.v_[2] },
	};

	dMatrix B(4,3, &_b[0][0]);
	B = B*0.5*dt;

	const double _q[3][3] = { 
		{1, 0, 0 },
		{0, 1, 0 },
		{0, 0, 1},
	};

	// Covariance matrix of porcess noises
	const dMatrix Q(3, 3, &_q[0][0]);

	// Predicted state estimate
	x = x * Quaternion (w[0]*dt, w[1]*dt, w[2]*dt);
	// Predicted estimate covariance
	P = A*P*~A + B*Q*~B;

	/////////////////////////////////////////////////////////////////////////
	// Update 

	//   ʴ Ȳ ׻ ߷ ӵ Ʒ (-z) ۿϰ ֱ
	// ,     ӵ (g) ߷° ϴ. ׷
	// ӵ  ڼκ  ڼ _Q   ִ.
	dVector g0 = dVector3(0., 0., -1.);
	Quaternion a (0, accl[0], accl[1], accl[2]);

	// g  ǥ踦   ̹Ƿ, ̵  ǥ  ǥϱ 
	// ؼ _Q ؼ  ǥ  ȯѴ.
	Quaternion ag = x * a * x.i();

	//  ǥ迡 ӵ   ߷ (Qg) ̻ ߷ (g0) ̷ 
	// normal vector(ng) հ(alpha) Ѵ.
	dVector ng = Cross (ag.v(), g0);
	double alpha = asin(Norm2(ng)/Norm2(ag.v()));

	dMatrix H = GetMeasurementMatrix (x);

	// ߷ ã   Ʈ ϴ (̵) Ѵ. 
	// ߷  ũⰡ 1.ó  ̵ Ŀϰ 1. ־  ̵  Ѵ.
	double l = sqrt (accl[0]*accl[0] + accl[1]*accl[1] + accl[2]*accl[2]) - 1.;
	double s = exp(-l*l);

	const double _r[1][1] = { 
		{ 1/s },
	};

	const dMatrix R(1, 1, &_r[0][0]);

	// Kalman gain
	dMatrix K = P*~H*!(H*P*~H + R);

	double Ka = K.normF ();

	// Updated state estimate
	x = Quaternion (Ka*alpha, ng) * x;
	x.unit();

	// Updated estimate covariance
	P = P - K*H*P;

	return x.RotationMatrix ();
}
